import logging
from typing import Optional

from numpy import ndarray
from pandas import Series
from pydantic import BaseModel, EmailStr

from ..schema import InferredField
from .base import BaseType

logger = logging.getLogger(__name__)


class Model(BaseModel):
    email: EmailStr


class EmailType(BaseType):
    weight = 5

    def validate(self, data: ndarray) -> Optional[InferredField]:
        for value in data:
            try:
                Model(email=value)
            except ValueError:
                logger.debug(f"Cannot cast the value '{value}' to {self.name}")
                return None
        return InferredField(inferred_type=self.name)

    def fix(self, column: Series, inferred_field: InferredField) -> Series:
        # Email defined as a virtual data-type. Fix is not relevant.
        raise NotImplementedError()


def get_type() -> EmailType:
    return EmailType()



from abc import ABC, abstractmethod
from typing import Optional

from numpy import ndarray
from pandas import Series

from ..schema import InferredField


class BaseType(ABC):
    weight: int = 0

    @abstractmethod
    def validate(self, data: ndarray) -> Optional[InferredField]:
        pass

    @abstractmethod
    def fix(self, column: Series, inferred_field: InferredField) -> Series:
        pass

    @property
    def name(self) -> str:
        return str(self.__class__.__name__.replace("Type", ""))



import re


def exception_router(exception: Exception):
    error_message = str(exception)
    match = re.match(
        r"(.*)Expected (.*) fields in line (.*), saw (.*)\n", error_message
    )
    if match is not None:
        raise WrongNumberOfColumnException(
            num_of_expected=match.group(2),
            num_of_saw=match.group(4),
            line_number=match.group(3),
        )

    match = re.match(
        r"(.*)data (.*) doesn't match format (.*), at position(.*)", error_message
    )
    if match is not None:
        raise FormatCastingException(
            bad_value=match.group(2), expected_format=match.group(3)
        )
    raise exception


class ShmessyException(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class WrongNumberOfColumnException(ShmessyException):
    def __init__(self, num_of_expected: int, num_of_saw: int, line_number: int):
        super().__init__(
            f"Error in line {line_number}: Expected {num_of_expected} fields, found {num_of_saw}."
        )


class FormatCastingException(ShmessyException):
    def __init__(self, bad_value: str, expected_format: str):
        super().__init__(
            f"The value {bad_value} doesn't match format {expected_format}."
        )



import logging
import math
from datetime import datetime
from enum import Enum
from typing import Optional

from numpy import ndarray
from pandas import Series, to_datetime

from ..schema import InferredField
from .base import BaseType

logger = logging.getLogger(__name__)


class TimestampResolution(str, Enum):
    SECONDS = "s"
    MILLISECONDS = "ms"
    NANOSECONDS = "ns"


class UnixTimestampType(BaseType):
    weight = 4
    min_valid_year: int = 1980
    max_valid_year: int = 2100

    @staticmethod
    def _unix_timestamp_resolution(value: int) -> TimestampResolution:
        number_of_digits = len(str(int(value)))
        if number_of_digits == 10:
            return TimestampResolution.SECONDS
        if number_of_digits == 13:
            return TimestampResolution.MILLISECONDS
        if number_of_digits == 16:
            return TimestampResolution.NANOSECONDS

    @staticmethod
    def _fix_input_resolution(
        value: int, selected_resolution: TimestampResolution
    ) -> int:
        if selected_resolution == TimestampResolution.SECONDS:
            return int(value)
        if selected_resolution == TimestampResolution.MILLISECONDS:
            return int(int(value) / 1000)
        if selected_resolution == TimestampResolution.NANOSECONDS:
            return int(int(value) / 1000 / 1000)

    @staticmethod
    def _is_non_numeric_float(value: float) -> bool:
        if isinstance(value, float) and (math.isnan(value) or math.isinf(value)):
            return True
        return False

    def validate(self, data: ndarray) -> Optional[InferredField]:
        try:
            selected_resolution = None
            for value in data:
                if not self._is_non_numeric_float(value):
                    if not selected_resolution:
                        selected_resolution = self._unix_timestamp_resolution(
                            int(value)
                        )
                        if not selected_resolution:
                            return None
                    parsed_value = datetime.utcfromtimestamp(
                        self._fix_input_resolution(value, selected_resolution)
                    )
                    if (
                        parsed_value.year < self.min_valid_year
                        or parsed_value.year > self.max_valid_year
                    ):
                        return None

            if not selected_resolution:
                return None

            return InferredField(
                inferred_type=self.name, inferred_pattern=selected_resolution
            )
        except ValueError as e:
            logger.debug(f"Cannot cast the given data to {self.name}: {e}")
            return None

    def fix(self, column: Series, inferred_field: InferredField) -> Series:
        return to_datetime(column, unit=inferred_field.inferred_pattern.value)


def get_type() -> UnixTimestampType:
    return UnixTimestampType()



from typing import Optional

from numpy import ndarray
from pandas import Series, to_datetime

from ..schema import InferredField
from . import validate_strptime_pattern
from .base import BaseType


class DatetimeType(BaseType):
    weight = 3
    patterns: list[str] = [
        "%d-%m-%Y %H:%M",  # 11-14-2003 00:00
        "%m/%d/%y %H:%M:%S",  # 12/15/22 00:00:00
        "%m-%d-%y %H:%M:%S",  # 12-30-2022 00:00:00
        "%m/%d/%Y %H:%M:%S",  # 12/30/2022 00:00:00
        "%m-%d-%Y %H:%M:%S",  # 12-30-2022 00:00:00
        "%Y/%m/%d %H:%M:%S",  # 2022/12/30 00:00:00
        "%Y-%m-%d %H:%M:%S",  # 2022-12-30 00:00:00
        "%Y-%m-%d %H:%M:%SZ",  # 2022-12-30 00:00:00Z
        "%Y-%m-%dT%H:%M:%SZ",  # 2022-12-30T00:00:00Z
        "%Y-%m-%dT%H:%M:%S.%f",  # 2022-12-30T00:00:00.000
        "%Y-%m-%d %H:%M:%S.%fZ",  # 2022-12-30 00:00:00.000Z
        "%Y-%m-%d %H:%M:%S.%f",  # 2022-12-30 00:00:00.000
        "%Y-%m-%dT%H:%M:%S.%fZ",  # 2022-12-30T00:00:00.000Z
        "%Y-%m-%dT%H:%M:%S",
    ]

    def validate(self, data: ndarray) -> Optional[InferredField]:
        for pattern in self.patterns:
            if validate_strptime_pattern(data, pattern):
                return InferredField(inferred_type=self.name, inferred_pattern=pattern)

    def fix(self, column: Series, inferred_field: InferredField) -> Series:
        return to_datetime(column, format=inferred_field.inferred_pattern)


def get_type() -> DatetimeType:
    return DatetimeType()
